/*
 * netinfofilterd.c - Netinfo Filter Daemon
 *
 * (c)2008,2009,2010,2018 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
*/


#include <sys/time.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <dirent.h>
#include <pthread.h>
#include <sys/inotify.h>

#include <pcre.h>

//#ifdef LUA_DISTRO_RPM
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
#ifndef lua_pushglobaltable
// For Lua versions prior to 5.2
#define lua_pushglobaltable(L)	lua_pushvalue(L, LUA_GLOBALSINDEX)
#endif

#include <dbi/dbi.h>

#include <cutil/slist.h>
#include <cutil/avltree.h>
#include <cutil/strfuncs.h>

#include <netinfo/netinfo.h>
#include <netinfo/confxml.h>



#define NETINFOFILTERD_CONFIG_FILE    "/netinfofilterd.conf"
#define NETINFOFILTERD_LOG_FILE       "/netinfofilterd.log"

#define	PATH_SIZE			256
#define	SQL_BUFFER_SIZE			1024
#define	LOG_BUFFER_SIZE			1024
#define TRAP_BUFFER_SIZE		16384
#define	UNIQUESTR_OID_SIZE		64
#define	STRING_BUFFER_SIZE		1024

#define	UNIQUESTR_FIELD_SIZE		64

#define	FLAG_FLUSH_LOG_ENTRIES		1
#define	FLAG_LOG_UNHANDLED_TRAPS	2
#define FLAG_NODAEMON			4

#define	CACHE_AGE_MAXIMUM		600
#define LOOP_INTERVAL_SECONDS		15

/* 16 PCRE ovector items */
#define	PCRE_OVEC_ITEMS			48

#define	RC_NOTFOUND			ENOENT
#define	RC_DBERROR			2001
#define	RC_CONTINUE			2002

#define	TF_FLAG_CONTINUE		1

#define	COMP_EQUAL			1
#define	COMP_NOT_EQUAL			2
#define	COMP_GREATERTHAN		3
#define	COMP_LESSTHAN			4
#define	COMP_GREATERTHAN_EQUAL		5
#define	COMP_LESSTHAN_EQUAL		6
#define	COMP_REGEX			7
#define	COMP_NOT_REGEX			8

#define VALTYPE_INTEGER			1
#define	VALTYPE_FLOAT			2
#define	VALTYPE_STRING			3

#define	STRPART_STRING			1
#define	STRPART_BACKREF			2

#define	MATCH_FLAG_PCREOID		1
#define	MATCH_FLAG_STRSTR		2

/*
#define	DEFAULT_TRAPOID			"snmpTrapOID.0"
*/
#define	DEFAULT_TRAPOID			"SNMPv2-MIB::snmpTrapOID.0"


typedef struct	_TrapOid		TrapOid;
struct	_TrapOid {
	char			*name;
	char			*value;
};


typedef struct	_Trap			Trap;
struct	_Trap {
	uint32_t		timestamp;
	char			*trapoid;
	char			*hostname;
	AVLTree			*tree;
	SList			*list;
};


typedef struct	_NVData			NVData;
struct	_NVData {
	TrapOid			spair;
	AVLTree			*tree;
};


typedef struct	_StrPart		StrPart;
struct	_StrPart {
	uint32_t		type;
	char			*string;
};


typedef struct	_FilterMatch	FilterMatch;
struct _FilterMatch {
	uint64_t	id;
	char		*oid;
	uint32_t	flags;
	uint32_t	comparator;
	uint32_t	valuetype;
	uint32_t	valuelen;
	pcre		*pcrecode;
	pcre		*pcreoid;
	union {
		uint64_t	ui64;
		int64_t		i64;
		float		f;
		double		d;
		char		*s;
		unsigned char	*us;
	} value;
};


typedef struct	_FilterRule	FilterRule;
struct _FilterRule {
	uint64_t		id;
	uint32_t		class;
	uint32_t		severity;
	uint32_t		uniqueid;
	uint32_t		markrecover;
	uint32_t		markunique;
	uint32_t		active;
	uint32_t		order;
	char			*name;
	char			*desc;
	char			*trapoid;
	char			*object;
	char			*msg;
	SList			*match_list;
	uint32_t		match_count;
	NVData			*actions;
	char			*uniquestr_oid;
	pcre			*uniquestr_regex;
	SList			*uniquestr_parts;

};


typedef	struct	_FilterListItem		FilterListItem;
struct	_FilterListItem {
	char			*oid;
	SList			*list;
};


typedef struct	_PerfData		PerfData;
struct	_PerfData {
	uint64_t		temp_ui64;
	uint64_t		sum_sec;
	uint64_t		sum_usec;
	uint64_t		sum_items;
	struct timeval		tv_start;
	struct timeval		tv_end;
};


typedef	struct	_ProcessState		ProcessState;
struct	_ProcessState {
	int			buffer_alloc;
	int			buffer_bytes;
	char			*buffer;
	int			matches;

	PerfData		perfdata;
};


typedef struct	_FilterManager	FilterManager;
struct _FilterManager {
	uint32_t		flags;

	uint32_t		loop_interval;
	uint32_t		loop_continue;

	uint32_t		cache_maxage;
	time_t			cache_lastload;
	SList			*cache_flilist;
	SList			*cache_rulelist;
	AVLTree			*cache_ruletree;
	pthread_rwlock_t	cache_rwlock;

	Configuration		*configuration;
	char			config_file[PATH_SIZE];

	dbi_conn		db_conn;
	char			*db_driver;
	char			*db_hostname;
	int			db_port;
	char			*db_dbname;
	char			*db_username;
	char			*db_password;

	lua_State		*lua_state;

	FILE			*log_fd;
	char			log_file[PATH_SIZE];
	pthread_mutex_t		log_mutex;
	char			log_buffer[LOG_BUFFER_SIZE];

	ProcessState		pstate;
	char			*default_trapoid;

	int			pickupdir_fd;
	char			*pickupdir_path;
	char			pickupdir_buffer[LOG_BUFFER_SIZE];
};



/* Function prototypes */
int db_connect (void);
int db_disconnect (void);
int db_ping (void);
int lua_initialize (void);
int lua_load_identifiers (void);
int lua_uninitialize (void);
int lua_execute_check (lua_State *ls, char *code);
int log_open (void);
int log_close (void);
int log_entry (char *format, ...);
void signal_handler (int sig);
int cache_refresh_rules (void);
int cache_unload_rules (void);
int process_trap_file (ProcessState *ps, char *trapfile);
int filter_rule_list_free (SList *list);
int filter_match_list_free (SList *list);
int filterlistitem_list_free (SList *list);
int filter_list_item_compare (FilterListItem *a, FilterListItem *b);
int filter_match_test (FilterMatch *fm, Trap *trap);
int trapoid_compare (TrapOid *a, TrapOid *b);
int trapoid_destroy (TrapOid *a);
TrapOid *trapoid_tree_find (AVLTree *tree, char *name);
int string_to_nvdata (NVData **nvdp,
			char *string,
			char *pairdelim,
			char nvdelim);
int handle_matching_trap (FilterRule *fr, Trap *trap);
int emit_notification (FilterRule *fr, Trap *trap, char *notifylist);
int parse_string_to_template (SList **slp, char *string);
int filter_rule_build_uniquestr (FilterRule *fr, Trap *trap, char **stringp);
int alert_add_data (uint64_t a_id, uint32_t type, char *message);
int load_configuration_file (char *file);
int trap_dump_to_string (Trap *trap, char *str, uint32_t strlen);
void usage(void);


/* Lua additional functions */
int lua_callback_alert_resolve (lua_State *ls);
int lua_callback_alert_acknowledge (lua_State *ls);
int lua_callback_emit_notification (lua_State *ls);
int lua_callback_alert_add_note (lua_State *ls);



/* Global variables */
FilterManager	Manager;
static const luaL_Reg lualibs[] = {
	{"base",	luaopen_base},
	{NULL,		NULL}
};



int main (int argc, char **argv)
{
	int			rc;
	int			bytes;
	pid_t			pid;
	int			option;
	struct timeval		tv;
	fd_set			rfds;
	char			path[LOG_BUFFER_SIZE];
	struct stat		st;
	char			*p = NULL;
	struct inotify_event	*ine = NULL;
	DIR			*dir = NULL;
	struct dirent		*de = NULL;

	// Set default values
	memset (&Manager, 0, sizeof (Manager));
	netinfo_get_component_filepath (Manager.config_file, PATH_SIZE, "etc", NETINFOFILTERD_CONFIG_FILE);
	netinfo_get_component_filepath (Manager.log_file, PATH_SIZE, "etc", NETINFOFILTERD_LOG_FILE);

	Manager.cache_maxage = CACHE_AGE_MAXIMUM;
	Manager.loop_interval = LOOP_INTERVAL_SECONDS;

	/* Parse commandline arguments */
	while (( option = getopt(argc, argv, "DFf:hl:V")) != EOF) {
		switch(option) {
			case 'D':
				Manager.flags |= FLAG_NODAEMON;
				fprintf (stderr, "Not daemonizing.\n");
				break;
			case 'F':
				Manager.flags |= FLAG_FLUSH_LOG_ENTRIES;
				break;
			case 'f':
				strncpy(Manager.config_file, optarg, PATH_SIZE);
				Manager.config_file[PATH_SIZE-1] = 0;
				break;
			case 'l':
				strncpy(Manager.log_file, optarg, PATH_SIZE);
				Manager.log_file[PATH_SIZE-1] = 0;
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'V':
				printf("\nnetinfofilterd - version %s\n\n",
					VERSION);
				exit(0);
				break;
			default:
				usage();
				exit(-1);
				break;
		}
	}

	rc = load_configuration_file (Manager.config_file);
	if (rc != 0) {
		fprintf (stderr, "Failed to load configuration file.\n");
		exit (rc);
	}

	rc = log_open ();
	if (rc != 0) {
		fprintf (stderr, "Failed to open log file.\n");
		exit (rc);
	}
	log_entry ("Netinfofilterd startup.\n");

	if (!(Manager.flags & FLAG_NODAEMON)) {
		if ( (pid=fork()) < 0)
			exit(-1);
		else if(pid != 0)
			exit(0);
		rc = chdir ("/");
		setsid ();
		umask (0);
		close (2);
		close (1);
		close (0);
	}

	rc = pthread_rwlock_init (&Manager.cache_rwlock, NULL);

	Manager.pstate.buffer = calloc (1, TRAP_BUFFER_SIZE);
	if (Manager.pstate.buffer == NULL) {
		log_entry ("Failed to allocate process state buffer.\n");
		log_entry ("Exiting.\n");
		log_close ();
		exit (rc);
	}
	Manager.pstate.buffer_alloc = TRAP_BUFFER_SIZE;
	Manager.pstate.buffer_bytes = 0;
	Manager.default_trapoid = DEFAULT_TRAPOID;

	log_entry ("Initializing inotify.\n");
	Manager.pickupdir_fd = inotify_init ();
	if (Manager.pickupdir_fd == -1) {
		rc = errno;
		log_entry ("Failed to initialize inotify. rc=%d(%s)\n",
			rc,
			strerror (rc));
		free (Manager.pstate.buffer);
		log_entry ("Exiting.\n");
		log_close ();
		exit (rc);
	}
	rc = inotify_add_watch (Manager.pickupdir_fd,
				Manager.pickupdir_path,
				IN_CLOSE_WRITE | IN_MOVED_TO);

	if (rc == -1) {
		rc = errno;
		log_entry ("Failed to add inotify watch dir. rc=%d(%s)\n",
			rc,
			strerror (rc));
		close (Manager.pickupdir_fd);
		free (Manager.pstate.buffer);
		log_entry ("Exiting.\n");
		log_close ();
		exit (rc);
	}

	log_entry ("Initializing lua environment.\n");
	rc = lua_initialize ();
	if (rc != 0) {
		close (Manager.pickupdir_fd);
		free (Manager.pstate.buffer);
		log_entry ("Failed to initialize Lua. rc=%d\n", rc);
		log_entry ("Exiting.\n");
		log_close ();
		exit (rc);
	}

	log_entry ("Initializing database connection.\n");
	rc = db_connect ();
	if (rc != 0) {
		log_entry ("Failed to initialize DB connection. rc=%d\n", rc);
		lua_uninitialize ();
		close (Manager.pickupdir_fd);
		free (Manager.pstate.buffer);
		log_entry ("Netinfofilterd exiting.\n");
		log_entry ("Exiting.\n");
		log_close ();
		exit (rc);
	}
	
	rc = lua_load_identifiers ();

	log_entry ("Startup completed.\n");
	Manager.loop_continue = 1;
	while (1) {

		if (!Manager.loop_continue)
			break;

		/* Load/update cache */
		rc = cache_refresh_rules ();

		/* wait for inotify events */
		FD_ZERO(&rfds);
		FD_SET(Manager.pickupdir_fd, &rfds);
		tv.tv_sec = Manager.loop_interval;
		tv.tv_usec = 0;
		rc = select (Manager.pickupdir_fd + 1,
			&rfds,
			NULL,
			NULL,
			&tv);
		if (rc > 0) {
			if (FD_ISSET (Manager.pickupdir_fd, &rfds)) {
				bytes = read (Manager.pickupdir_fd,
						Manager.pickupdir_buffer,
						LOG_BUFFER_SIZE);
				if (bytes > 0) {
					p = Manager.pickupdir_buffer;
					while (bytes > 0) {
						ine = (struct inotify_event *)p;
						/* Process trap file */
						log_entry ("Trap file found: %s\n",
							ine->name);
						strcpy (path, Manager.pickupdir_path);
						strcat (path, "/");
						strcat (path, ine->name);
						rc = process_trap_file (&Manager.pstate,
									 path);
						rc = unlink (path);
						bytes -= sizeof (struct inotify_event) + ine->len;
						p += sizeof (struct inotify_event) + ine->len;
						
					}
				} else if (bytes == -1) {
					log_entry ("inotify read(): rc=%d(%s)\n",
						rc,
						strerror (rc));
				}
			}
		} else if (rc == 0) {
			/* timeout, scan the directory */
			dir = opendir (Manager.pickupdir_path);
			if (dir != NULL) {
				while ( (de = readdir (dir)) != NULL) {
					strcpy (path, Manager.pickupdir_path);
					strcat (path, "/");
					strcat (path, de->d_name);
					rc = stat (path, &st);
					if (S_ISREG(st.st_mode)) {
						log_entry ("Trap file found: %s\n",
							de->d_name);
						rc = process_trap_file (&Manager.pstate,
									 path);
						rc = unlink (path);
					}
				}
				closedir(dir);
			} else {
				rc = errno;
				log_entry ("opendir(): rc=%d(%s)\n",
					rc,
					strerror (rc));
				break;
			}
		} else {
			rc = errno;
			log_entry ("select(): rc=%d(%s)\n",
				rc,
				strerror (rc));
			break;
		}
	}

	rc = db_disconnect ();
	rc = lua_uninitialize ();
	close (Manager.pickupdir_fd);
	free (Manager.pstate.buffer);
	log_entry ("Netinfofilterd shutdown.\n");
	rc = log_close ();
	return rc;
}


int db_connect (void)
{
	int	rc;
	const char	*errmsg = NULL;
	if (Manager.db_conn != NULL)
		return EINVAL;
	dbi_initialize (NULL);
	Manager.db_conn = dbi_conn_new (Manager.db_driver);
	if (Manager.db_conn == NULL) {
		log_entry ("Failed to allocate a DB connection structure.\n");
		return EINVAL;
	}
	rc = dbi_conn_set_option (Manager.db_conn,
				"host",
				Manager.db_hostname);
	rc = dbi_conn_set_option_numeric (Manager.db_conn,
				"port",
				Manager.db_port);
	rc = dbi_conn_set_option (Manager.db_conn,
				"username",
				Manager.db_username);
	rc = dbi_conn_set_option (Manager.db_conn,
				"password",
				Manager.db_password);
	rc = dbi_conn_set_option (Manager.db_conn,
				"dbname",
				Manager.db_dbname);
	rc = dbi_conn_connect (Manager.db_conn);
	if (rc != 0) {
		rc = dbi_conn_error (Manager.db_conn, &errmsg);
		log_entry ("db_conn_connect(): rc=%d(%s)\n", rc, errmsg);
		dbi_conn_close (Manager.db_conn);	
		Manager.db_conn = NULL;
		return rc;
	}
	log_entry ("Successfully connected to the database.\n");
	return 0;
}


int db_disconnect (void)
{
	if (Manager.db_conn == NULL) {
		log_entry ("Attempted to close an unconnect DB connection.\n");
		return EINVAL;
	}
	dbi_conn_close (Manager.db_conn);
	Manager.db_conn = NULL;
	dbi_shutdown ();
	log_entry ("Successfully disconnected from the database.\n");
	return 0;
}


int lua_initialize (void)
{
	lua_State	*ls = NULL;
	const luaL_Reg	*llib = NULL;

	/* Create lua state */
	ls = luaL_newstate ();
	if (ls == NULL) {
		log_entry ("Failed to initialize lua state.\n");
		return 1;
	}

	/* Load the extension libraries */
	for (llib = lualibs ; llib->func != NULL ; llib++) {
		llib->func (ls);
		lua_settop (ls, 0);
	}

	// get global environment table from registry
	lua_pushglobaltable(ls);

	/* Register the Lua functions */
	lua_pushstring (ls, "alert_resolve");
	lua_pushcfunction (ls, lua_callback_alert_resolve);
	//lua_settable (ls, LUA_GLOBALSINDEX);
	lua_settable (ls, -3);

	lua_pushstring (ls, "alert_acknowledge");
	lua_pushcfunction (ls, lua_callback_alert_acknowledge);
	//lua_settable (ls, LUA_GLOBALSINDEX);
	lua_settable (ls, -3);

	lua_pushstring (ls, "emit_notification");
	lua_pushcfunction (ls, lua_callback_emit_notification);
	//lua_settable (ls, LUA_GLOBALSINDEX);
	lua_settable (ls, -3);

	lua_pushstring (ls, "alert_add_note");
	lua_pushcfunction (ls, lua_callback_alert_add_note);
	//lua_settable (ls, LUA_GLOBALSINDEX);
	lua_settable (ls, -3);

	// dont remove the global environment table from the stack
	// We'll use it later
	//lua_pop(pLuaState, 1);

	Manager.lua_state = ls;
	log_entry ("Successfully initialized the Lua environment.\n");
	return 0;
}


int lua_load_identifiers (void)
{
	int		count;
	dbi_result	res;
	int		id;
	const char	*name = NULL;
	lua_State	*ls = NULL;
	if (Manager.db_conn == NULL || Manager.lua_state == NULL)
		return EINVAL;
	ls = Manager.lua_state;

	/* Load alert state names/ids */
	count = 0 ;
	res = dbi_conn_query (Manager.db_conn,
				"select ast_id,ast_name from alert_state");
	if (res != NULL) {
		while (dbi_result_next_row (res)) {
			id = dbi_result_get_uint (res, "ast_id");
			name = dbi_result_get_string (res, "ast_name");
			lua_pushstring (ls, name);
			lua_pushnumber (ls, id);
			//lua_settable (ls, LUA_GLOBALSINDEX);
			lua_settable (ls, -3);
			count++;
		}
		dbi_result_free (res);
	}
	log_entry ("Loaded %d alert state identifier(s).\n", count);

	/* Load alert severity names/ids */
	count = 0 ;
	res = dbi_conn_query (Manager.db_conn,
				"select as_id,as_name from alert_severity");
	if (res != NULL) {
		while (dbi_result_next_row (res)) {
			id = dbi_result_get_uint (res, "as_id");
			name = dbi_result_get_string (res, "as_name");
			lua_pushstring (ls, name);
			lua_pushnumber (ls, id);
			//lua_settable (ls, LUA_GLOBALSINDEX);
			lua_settable (ls, -3);
			count++;
		}
		dbi_result_free (res);
	}
	log_entry ("Loaded %d alert severity identifier(s).\n", count);

	/* Load alert class names/ids */
	count = 0 ;
	res = dbi_conn_query (Manager.db_conn,
				"select ac_id,ac_name from alert_class");
	if (res != NULL) {
		while (dbi_result_next_row (res)) {
			id = dbi_result_get_uint (res, "ac_id");
			name = dbi_result_get_string (res, "ac_name");
			lua_pushstring (ls, name);
			lua_pushnumber (ls, id);
			//lua_settable (ls, LUA_GLOBALSINDEX);
			lua_settable (ls, -3);
			count++;
		}
		dbi_result_free (res);
	}
	log_entry ("Loaded %d alert class identifier(s).\n", count);
	
	return 0;
}


int lua_uninitialize (void)
{
	lua_close (Manager.lua_state);
	log_entry ("Successfully uninitialized the Lua environment.\n");
	return 0;
}


int lua_execute_check (lua_State *ls, char *code)
{
	int		rc;
	if (ls == NULL || code == NULL)
		return EINVAL;

	/* TODO */

	/* Load the code(text/binary) to execute */
	rc = luaL_loadbuffer (ls, code, strlen (code), "alert_check");
	if (rc != 0) {
		switch (rc) {
			case LUA_ERRSYNTAX:
				log_entry ("A Lua syntax error occurred.\n");
				break;
			case LUA_ERRMEM:
				log_entry ("A memory allocation error occurred.\n");
				break;
		}
		return rc;
	}
	/* Execute the loaded code */
	rc = lua_pcall (ls, 0, LUA_MULTRET, 0);
	if (rc != 0) {
		switch (rc) {
			case LUA_ERRRUN:
				log_entry ("A Lua runtime error occurred.\n");
				break;
			case LUA_ERRMEM:
				log_entry ("A memory allocation error occurred.\n");
				break;
		}
	}

	/* TODO */

	return rc;
}


int log_open (void)
{
	int	rc;
	if (Manager.log_file == NULL)
		return EINVAL;
	Manager.log_fd = fopen (Manager.log_file, "a");
	if (Manager.log_fd == NULL)
		return errno;
	rc = pthread_mutex_init (&Manager.log_mutex, NULL);
	if (rc != 0) {
		fclose (Manager.log_fd);
	}
	return rc;
}


int log_close (void)
{
	int	rc;
	fclose (Manager.log_fd);
	Manager.log_fd = NULL;
	rc = pthread_mutex_destroy (&Manager.log_mutex);
	return rc;
}


int log_entry (char *format, ...)
{
	va_list		args;
	int		rc;
	time_t		t;
	struct tm	*lt = NULL;
	if (format == NULL || Manager.log_fd == NULL)
		return EINVAL;
	va_start (args, format);
	t = time (NULL);
	lt = localtime (&t);
	pthread_mutex_lock (&Manager.log_mutex);
	fprintf (Manager.log_fd,
		"%04d-%02d-%02d %02d:%02d:%02d ",
		lt->tm_year+1900,
		lt->tm_mon+1,
		lt->tm_mday,
		lt->tm_hour,
		lt->tm_min,
		lt->tm_sec);
	rc = vsnprintf (Manager.log_buffer,
			LOG_BUFFER_SIZE,
			format,
			args);
	fprintf (Manager.log_fd, "%s", Manager.log_buffer);
	pthread_mutex_unlock (&Manager.log_mutex);
	if (Manager.flags & FLAG_FLUSH_LOG_ENTRIES)
		fflush (Manager.log_fd);
	va_end (args);
	return 0;
}


void signal_handler (int sig)
{
	switch (sig) {
		case SIGINT:
		case SIGTERM:
		case SIGQUIT:
			Manager.loop_continue = 0;
			break;
	}
	return;
}


int cache_refresh_rules (void)
{
	int			rc;
	int			count;
	int			pcreerr;
	time_t			t;
	dbi_result		result;
	char			sql[SQL_BUFFER_SIZE];
	FilterListItem		flis;
	const char		*pcreerrstr = NULL;
	const char		*comp = NULL;
	const char		*value = NULL;
	const char		*errmsg = NULL;
	SList			*litem = NULL;
	SList			*newlist = NULL;
	SList			*oldlist = NULL;
	SList			*newflilist = NULL;
	SList			*oldflilist = NULL;
	AVLTree			*newtree = NULL;
	AVLTree			*oldtree = NULL;
	FilterRule		*fr = NULL;
	FilterMatch		*fm = NULL;
	FilterListItem		*fli = NULL;
	TrapOid			*to = NULL;

	/*
	 * Check if the cache is still valid
	*/
	t = time (NULL);
	pthread_rwlock_rdlock (&Manager.cache_rwlock);
	if ( (t - Manager.cache_lastload) < Manager.cache_maxage) {
		pthread_rwlock_unlock (&Manager.cache_rwlock);
		return 0;
	}
	pthread_rwlock_unlock (&Manager.cache_rwlock);

	flis.oid = NULL;
	flis.list = NULL;

	/*
	 * Create a new tree
	*/
	rc = avltree_new (&newtree,
			(AVLNodeCompare)filter_list_item_compare,
			NULL);
	if (rc != 0) {
		log_entry ("Failed to allocate avltree.\n");
		return rc;
	}

	/*
	 * Create a new rule list from the database
	*/
	result = dbi_conn_query (Manager.db_conn, 
				"select tf_id,tf_class,tf_severity,tf_uniqueid,tf_markrecover,tf_name,tf_desc,tf_trapoid,tf_object,tf_msg,tf_unique,tf_order from trap_filter order by tf_trapoid,tf_order asc");
	if (result == NULL) {
		rc = dbi_conn_error (Manager.db_conn, &errmsg);
		log_entry ("Failed to execute filter rule SQL.\n");
		log_entry ("dbierr: %s\n", errmsg);
		avltree_free (&newtree);
		return RC_DBERROR;
	}
	count = 0;
	while (dbi_result_next_row (result)) {
		fr = calloc (1, sizeof (FilterRule));
		if (fr == NULL) {
			log_entry ("Failed to allocate filter rule memory.\n");
			dbi_result_free (result);
			filter_rule_list_free (newlist);
			avltree_free (&newtree);
			return ENOMEM;
		}
		fr->active = 1;
		fr->id = dbi_result_get_ulonglong (result, "tf_id");
		fr->class = dbi_result_get_uint (result, "tf_class");
		fr->severity = dbi_result_get_uint (result, "tf_severity");
		fr->uniqueid = dbi_result_get_uint (result, "tf_uniqueid");
		fr->markrecover = dbi_result_get_uint (result, "tf_markrecover");
		fr->markunique = dbi_result_get_uint (result, "tf_unique");
		fr->order = dbi_result_get_uint (result, "tf_order");
		fr->name = strdup(dbi_result_get_string (result, "tf_name"));
		fr->desc = strdup(dbi_result_get_string (result, "tf_desc"));
		fr->trapoid = strdup(dbi_result_get_string (result, "tf_trapoid"));
		fr->object = strdup(dbi_result_get_string (result, "tf_object"));
		fr->msg = strdup(dbi_result_get_string (result, "tf_msg"));

		rc = string_to_nvdata (&fr->actions, fr->msg, ";", '=');
		to = trapoid_tree_find (fr->actions->tree, "uniquestr_oid");
		if (to != NULL)
			fr->uniquestr_oid = strdup (to->value);
		to = trapoid_tree_find (fr->actions->tree, "uniquestr_regex");
		if (to != NULL) {
			fr->uniquestr_regex = pcre_compile (to->value,
							0,
							&pcreerrstr,
							&pcreerr,
							NULL);
		}
		to = trapoid_tree_find (fr->actions->tree, "uniquestr");
		if (to != NULL) {
			rc = parse_string_to_template (&fr->uniquestr_parts,
							to->value);
		}

		if (fr->name == NULL) {
			log_entry ("Failed to allocate filter rule memory.\n");
			dbi_result_free (result);
			free (fr);
			filter_rule_list_free (newlist);
			avltree_free (&newtree);
			return ENOMEM;
		}
		rc = slist_item_add (&newlist, fr);
		count++;
	}
	dbi_result_free (result);

	/*
	 * Load the filter matches into each rule
	*/
	litem = newlist;
	while (litem != NULL) {
		fr = slist_get_data (litem);
		rc = snprintf (sql,
				SQL_BUFFER_SIZE,
				"select tfm_id,tfm_tf_id,tfm_flags,tfm_valuetype,tfm_oid,tfm_comparator,tfm_value from trap_filter_match where tfm_tf_id = %lu order by tfm_order asc",
				fr->id);
		result = dbi_conn_query (Manager.db_conn, sql);
		if (result == NULL) {
			rc = dbi_conn_error (Manager.db_conn, &errmsg);
			log_entry ("Failed to execute filter match SQL.\n");
			log_entry ("dbierr: %s\n", errmsg);
			filter_rule_list_free (newlist);
			avltree_free (&newtree);
			return RC_DBERROR;
		}
		while (dbi_result_next_row (result)) {
			fm = calloc (1, sizeof (FilterMatch));
			if (fm == NULL) {
				log_entry ("Failed to allocate filter match memory.\n");
				dbi_result_free (result);
				filter_rule_list_free (newlist);
				avltree_free (&newtree);
				return ENOMEM;
			}
			fm->id = dbi_result_get_ulonglong (result, "tfm_id");
			fm->flags = dbi_result_get_uint(result, "tfm_flags");
			fm->valuetype = dbi_result_get_uint(result, "tfm_valuetype");
			fm->oid = strdup(dbi_result_get_string (result, "tfm_oid"));
			comp = dbi_result_get_string (result, "tfm_comparator");
			fm->valuelen = dbi_result_get_field_length (result, "tfm_value");
			value = dbi_result_get_string (result, "tfm_value");

			if (fm->flags & MATCH_FLAG_PCREOID) {
				fm->pcreoid = pcre_compile (fm->oid,
								0,
								&pcreerrstr,
								&pcreerr,
								NULL);
				if (fm->pcreoid == NULL) {
					log_entry ("Error: tfm_id=%llu Failed to compile regex '%s' %d(%s)\n",
						fm->id,
						fm->oid,
						pcreerr,
						pcreerrstr);
					fm->flags -= (fm->flags & MATCH_FLAG_PCREOID);
				}
			}

			if (!strcmp (comp, "=")) {
				fm->comparator = COMP_EQUAL;
			} else if (!strcmp (comp, "!=")) {
				fm->comparator = COMP_NOT_EQUAL;
			} else if (!strcmp (comp, ">")) {
				fm->comparator = COMP_GREATERTHAN;
			} else if (!strcmp (comp, "<")) {
				fm->comparator = COMP_LESSTHAN;
			} else if (!strcmp (comp, ">=")) {
					fm->comparator = COMP_GREATERTHAN_EQUAL;
			} else if (!strcmp (comp, "<=")) {
				fm->comparator = COMP_LESSTHAN_EQUAL;
			} else if (!strcmp (comp, "regex")) {
				fm->comparator = COMP_REGEX;
				fm->pcrecode = pcre_compile (value,
								0,
								&pcreerrstr,
								&pcreerr,
								NULL);
				if (fm->pcrecode == NULL) {
					log_entry ("Error: tfm_id=%llu Failed to compile regex '%s' %d(%s)\n",
						fm->id,
						value,
						pcreerr,
						pcreerrstr);
				}
			} else if (!strcmp (comp, "!regex")) {
				fm->comparator = COMP_NOT_REGEX;
				fm->pcrecode = pcre_compile (value,
								0,
								&pcreerrstr,
								&pcreerr,
								NULL);
				if (fm->pcrecode == NULL) {
					log_entry ("Error: tfm_id=%llu Failed to compile regex '%s' %d(%s)\n",
						fm->id,
						value,
						pcreerr,
						pcreerrstr);
				}
			} else {
				log_entry ("Error: invalid comperator type '%s' in match id %llu.\n",
					comp,
					fm->id);
			}
			if (fm->valuetype == VALTYPE_INTEGER) {
				fm->value.i64 = atoi(value);
			} else if (fm->valuetype == VALTYPE_FLOAT) {
				fm->value.f = atof(value);
			} else if (fm->valuetype == VALTYPE_STRING) {
				fm->value.s = strdup(value);
			} else {
				log_entry ("Error: invalid value type '%d' in match id %llu.\n",
					fm->valuetype,
					fm->id);
			}
			if (fm->oid == NULL) {
				log_entry ("Failed to allocate filter match memory.\n");
				dbi_result_free (result);
				free (fm);
				filter_rule_list_free (newlist);
				avltree_free (&newtree);
				return ENOMEM;
			}
			rc = slist_item_add (&fr->match_list, fm);
			fr->match_count++;
		}
		dbi_result_free (result);

		/*
		 * Add the filter rule to the new tree
		*/
		flis.oid = fr->trapoid;
		fli = avltree_find (newtree, &flis);
		if (fli != NULL) {
			slist_item_add (&fli->list, fr);
		} else {
			fli = calloc (1, sizeof (FilterListItem));
			if (fli != NULL) {
				fli->oid = strdup (fr->trapoid);
				slist_item_add (&fli->list, fr);
				avltree_insert (newtree, fli);
				slist_item_add (&newflilist, fli);
			} else {
				log_entry ("Failed to allocate filter list item memory.\n");
			}
		}

		litem = slist_get_next (litem);
	}

	/*
	 * Replace old rule list/tree with new list/tree
	*/
	pthread_rwlock_wrlock (&Manager.cache_rwlock);
	oldlist = Manager.cache_rulelist;
	oldtree = Manager.cache_ruletree;
	oldflilist = Manager.cache_flilist;
	Manager.cache_rulelist = newlist;
	Manager.cache_ruletree = newtree;
	Manager.cache_flilist = newflilist;
	Manager.cache_lastload = t;
	pthread_rwlock_unlock (&Manager.cache_rwlock);
	log_entry ("Refreshed rule cache (%d rule(s)).\n", count);

	/*
	 * Free the old list/tree
	*/
	rc = filter_rule_list_free (oldlist);
	rc = avltree_free (&oldtree);
	rc = filterlistitem_list_free (oldflilist);

	return 0;
}


int cache_unload_rules (void)
{
	int		rc;
	SList		*rulelist = NULL;
	SList		*flilist = NULL;
	AVLTree		*ruletree = NULL;

	/*
	 * Remove the rule list
	*/
	pthread_rwlock_wrlock (&Manager.cache_rwlock);
	rulelist = Manager.cache_rulelist;
	ruletree = Manager.cache_ruletree;
	flilist = Manager.cache_flilist;
	Manager.cache_rulelist = NULL;
	Manager.cache_lastload = 0;
	pthread_rwlock_unlock (&Manager.cache_rwlock);

	/*
	 * Free the list
	*/
	rc = filter_rule_list_free (rulelist);
	rc = avltree_free (&ruletree);
	rc = filterlistitem_list_free (flilist);

	return 0;
}


int process_trap_file (ProcessState *ps, char *trapfile)
{
	int			rc;
	int			count;
	int			items;
	int			match_count;
	int			trap_handled;
	FilterListItem		flis;
	TrapOid			strapoid;
	FILE			*fd = NULL;
	char			*p = NULL;
	char			**nvpair = NULL;
	FilterListItem		*tffli = NULL;
	SList			*tfitem = NULL;
	SList			*tfmitem = NULL;
	SList			*titem = NULL;
	FilterRule		*fr = NULL;
	FilterMatch		*fm = NULL;
	Trap			*trap = NULL;
	TrapOid			*trapoid = NULL;

	/*
	 * Performance Counters
	*/
	gettimeofday (&ps->perfdata.tv_start, NULL);

	/*
	 * Allocate the trap resources
	*/
	trap = calloc (1, sizeof (Trap));
	if (trap == NULL) {
		log_entry ("Failed to allocate trap structure.\n");
		return ENOMEM;
	}
	rc = avltree_new (&trap->tree,
			(AVLNodeCompare)trapoid_compare,
			(AVLNodeDestroy)trapoid_destroy);
	if (rc != 0) {
		log_entry ("Failed to allocate trap->tree structure.\n");
		free (trap);
		return ENOMEM;
	}

	/*
	 * Assign defaults
	*/
	trap->timestamp = time (NULL);

	/*
	 * Read in the trap file
	*/
	fd = fopen (trapfile, "r");
	if (fd == NULL) {
		log_entry ("Failed to open trap file '%s'.\n",
			trapfile);
		free (trap);
		return ENOENT;
	}
	while (!feof (fd)) {
		fgets (Manager.pstate.buffer,
			Manager.pstate.buffer_alloc,
			fd);
		items = 3;
		rc = strnsplit (Manager.pstate.buffer,
				strlen (Manager.pstate.buffer),
				' ',
				&nvpair,
				&items);
		if (items == 3) {
			count = strlen (nvpair[2]);
			if (count > 1) {
				p = nvpair[2];
			 	p[count-1] = 0;
			}
			free (nvpair[0]);
			nvpair[0] = nvpair[1];
			nvpair[1] = nvpair[2];
			nvpair[2] = NULL;
			/* HACK: cast nvpair to a trapOid */
			trapoid = (TrapOid *)nvpair;
			avltree_insert (trap->tree, trapoid);
			slist_item_add (&trap->list, trapoid);
		} else if (items == 2) {
			count = strlen (nvpair[1]);
			if (count > 1) {
				p = nvpair[1];
			 	p[count-1] = 0;
			}
			if (!strcmp (nvpair[0], "HOST:")) {
				trap->hostname = nvpair[1];
				free (nvpair[0]);
				free (nvpair);
			} else if (!strcmp (nvpair[0], "TIMESTAMP:")) {
				trap->timestamp = atoi (nvpair[1]);
			} else {
				/* TODO: append value to the prior oid */
			}
		}
	}
	fclose (fd);
	strapoid.name = Manager.default_trapoid;
	strapoid.value = NULL;
	trapoid = avltree_find (trap->tree, &strapoid);
	if (trapoid == NULL) {
		log_entry ("trap is missing the trapoid!\n");
		rc = slist_free (&trap->list);
		rc = avltree_free (&trap->tree);
		if (trap->trapoid != NULL)
			free (trap->trapoid);
		if (trap->hostname != NULL)
			free (trap->hostname);
		free (trap);
		return RC_NOTFOUND;
	}
	trap->trapoid = strdup (trapoid->value);

	/*
	 * Retrieve the SList of filters from the trapoid tree
	*/
	flis.oid = trap->trapoid;
	flis.list = NULL;
	pthread_rwlock_rdlock (&Manager.cache_rwlock);
	tffli = avltree_find (Manager.cache_ruletree, &flis);

	/*
	 * Test the trap against each filter rule
	*/
	trap_handled = 0;
	if (tffli) {
		tfitem = tffli->list;
		while (tfitem != NULL) {
			fr = slist_get_data (tfitem);
	
			/*
			 * Test each filter match in this filter rule
			*/
			match_count = 0;
			tfmitem = fr->match_list;
			while (tfmitem != NULL) {
				fm = slist_get_data (tfmitem);
				rc = filter_match_test (fm, trap);
				if (rc == 1)
					match_count++;
				tfmitem = slist_get_next (tfmitem);
			}
	
			if (match_count == fr->match_count) {
				rc = handle_matching_trap (fr, trap);
				trap_handled++;
				if (rc != RC_CONTINUE)
					break;
			}
			tfitem = slist_get_next (tfitem);
		}
	} else {
		log_entry ("Unhandled trap: No trap filters were found for this OID\n");
	}
	pthread_rwlock_unlock (&Manager.cache_rwlock);

	/*
	 * Log all unhandled traps
	*/
	if (trap_handled == 0 && (Manager.flags & FLAG_LOG_UNHANDLED_TRAPS)) {
		log_entry ("****************************************\n");
		log_entry ("Unhandled trap:\n");
		log_entry ("Hostname: %s\n", trap->hostname);
		log_entry ("TrapOID: %s\n", trap->trapoid);
		titem = trap->list;
		while (titem != NULL) {
			trapoid = slist_get_data (titem);
			log_entry ("%s: %s\n", trapoid->name, trapoid->value);
			titem = slist_get_next (titem);
		}
		log_entry ("\n");
	}

	/*
	 * Free the trap resources
	*/
	rc = slist_free (&trap->list);
	rc = avltree_free (&trap->tree);
	if (trap->trapoid != NULL)
		free (trap->trapoid);
	if (trap->hostname != NULL)
		free (trap->hostname);
	free (trap);

	/*
	 * Performance Counters
	*/
	gettimeofday (&ps->perfdata.tv_end, NULL);
	ps->perfdata.temp_ui64 += (uint64_t)(ps->perfdata.tv_end.tv_sec - 
						ps->perfdata.tv_start.tv_sec) * 1000000;
	ps->perfdata.temp_ui64 += (uint64_t)ps->perfdata.tv_end.tv_usec - ps->perfdata.tv_start.tv_usec;
	ps->perfdata.sum_items++;
	if (ps->perfdata.sum_items == 100) {
		log_entry ("PerfData: process_trap_file = %llu usec\n",
				ps->perfdata.temp_ui64);
		log_entry ("PerfData: process_trap_file = %0.3f usec/item\n",
				(float)ps->perfdata.temp_ui64 / (float)ps->perfdata.sum_items);
		memset (&ps->perfdata, 0, sizeof (PerfData));
	}

	return 0;
}


int filter_rule_list_free (SList *list)
{
	SList		*litem = NULL;
	FilterRule	*fr = NULL;
	if (list == NULL)
		return EINVAL;
	litem = list;
	while (litem != NULL) {
		fr = slist_get_data (litem);
		if (fr->name)
			free (fr->name);
		if (fr->desc)
			free (fr->desc);
		if (fr->trapoid)
			free (fr->trapoid);
		if (fr->object)
			free (fr->object);
		if (fr->msg)
			free (fr->msg);
		if (fr->actions) {
			avltree_free (&fr->actions->tree);
			free (fr->actions);
		}
		if (fr->uniquestr_oid)
			free (fr->uniquestr_oid);
		if (fr->uniquestr_regex)
			pcre_free (fr->uniquestr_regex);
		filter_match_list_free (fr->match_list);
		free (fr);
		litem = slist_get_next (litem);
	}
	slist_free (&list);
	return 0;
}


int filter_match_list_free (SList *list)
{
	SList		*litem = NULL;
	FilterMatch	*fm = NULL;
	if (list == NULL)
		return EINVAL;
	litem = list;
	while (litem != NULL) {
		fm = slist_get_data (litem);
		if (fm->oid)
			free (fm->oid);
		if (fm->valuetype == VALTYPE_STRING) {
			if (fm->value.s)
				free (fm->value.s);
		}
		free (fm);
		litem = slist_get_next (litem);
	}
	slist_free (&list);
	return 0;
}


int filterlistitem_list_free (SList *list)
{
	SList		*litem = NULL;
	FilterListItem	*fli = NULL;
	if (list == NULL)
		return EINVAL;
	litem = list;
	while (litem != NULL) {
		fli = slist_get_data (litem);
		if (fli->oid)
			free (fli->oid);
		slist_free (&fli->list);
		free (fli);
		litem = slist_get_next (litem);
	}
	slist_free (&list);
	return 0;
}


int lua_callback_alert_resolve (lua_State *ls)
{
	int		args;
	long long	a_id;
	lua_Number	n;
	dbi_result	result;
	const char	*str = NULL;
	args = lua_gettop (ls);
	log_entry ("alert_resolve():\n");
	if (!lua_isstring (ls, 1)) {
		log_entry ("alert_resolve: Argument is not a string.\n");
		return 0;
	}
	str = lua_tostring (ls, 1);

	lua_getglobal (ls, "alert");
	lua_pushstring (ls, "id");
	lua_gettable (ls, -2);
	n = lua_tonumber (ls, -1);
	a_id = (long long)n;

	result = dbi_conn_query (Manager.db_conn, "start transaction");
	if (result == NULL) {
		log_entry ("Failed to execute alert_resolve start tx SQL.\n");
		return 0;
	}
	dbi_result_free (result);

	result = dbi_conn_queryf (Manager.db_conn, 
				"insert into alert_data (ad_a_id,ad_timestamp,ad_type,ad_name,ad_data) values (%llu,NOW(),5,'SYSTEM','%s')",
				a_id,
				str);
	if (result == NULL) {
		log_entry ("Failed to execute alert_resolve data SQL (rollback).\n");
		result = dbi_conn_query (Manager.db_conn, "rollback");
		dbi_result_free (result);
		return 0;
	}
	dbi_result_free (result);
	result = dbi_conn_queryf (Manager.db_conn, 
				"update alert set a_state = 5 where a_id = %llu",
				a_id);
	if (result == NULL) {
		log_entry ("Failed to execute alert_resolve alert SQL (rollback).\n");
		result = dbi_conn_query (Manager.db_conn, "rollback");
		dbi_result_free (result);
		return 0;
	}
	dbi_result_free (result);

	result = dbi_conn_query (Manager.db_conn, "commit");
	if (result == NULL) {
		log_entry ("Failed to execute alert_resolve commit tx SQL.\n");
		return 0;
	}
	dbi_result_free (result);

	return 0;
}


int lua_callback_alert_acknowledge (lua_State *ls)
{
	int		args;
	long long	a_id;
	lua_Number	n;
	dbi_result	result;
	const char	*str = NULL;
	args = lua_gettop (ls);
	if (!lua_isstring (ls, 1)) {
		log_entry ("%s: Argument is not a string.\n",
			__PRETTY_FUNCTION__);
		return 0;
	}
	str = lua_tostring (ls, 1);

	lua_getglobal (ls, "alert");
	lua_pushstring (ls, "id");
	lua_gettable (ls, -2);
	n = lua_tonumber (ls, -1);
	a_id = (long long)n;

	result = dbi_conn_query (Manager.db_conn, "start transaction");
	if (result == NULL) {
		log_entry ("%s: Failed to execute start tx SQL.\n",
			__PRETTY_FUNCTION__);
		return 0;
	}
	dbi_result_free (result);

	result = dbi_conn_queryf (Manager.db_conn, 
				"insert into alert_data (ad_a_id,ad_timestamp,ad_type,ad_name,ad_data) values (%llu,NOW(),3,'SYSTEM','%s')",
				a_id,
				str);
	if (result == NULL) {
		log_entry ("%s: Failed to execute data SQL (rollback).\n",
			__PRETTY_FUNCTION__);
		result = dbi_conn_query (Manager.db_conn, "rollback");
		dbi_result_free (result);
		return 0;
	}
	dbi_result_free (result);
	result = dbi_conn_queryf (Manager.db_conn, 
				"update alert set a_state = 3 where a_id = %llu",
				a_id);
	if (result == NULL) {
		log_entry ("%s: Failed to execute alert SQL (rollback).\n",
			__PRETTY_FUNCTION__);
		result = dbi_conn_query (Manager.db_conn, "rollback");
		dbi_result_free (result);
		return 0;
	}
	dbi_result_free (result);

	result = dbi_conn_query (Manager.db_conn, "commit");
	if (result == NULL) {
		log_entry ("%s: Failed to execute commit tx SQL.\n",
			__PRETTY_FUNCTION__);
		return 0;
	}
	dbi_result_free (result);
	return 0;
}


int lua_callback_emit_notification (lua_State *ls)
{
	int		args;
	const char	*str = NULL;
	args = lua_gettop (ls);
	str = lua_tostring (ls, 1);
	log_entry ("emit_notification(): str='%s'\n", str);
	return 0;
}


int lua_callback_alert_add_note (lua_State *ls)
{
	int		args;
	long long	a_id;
	lua_Number	n;
	dbi_result	result;
	const char	*str = NULL;
	args = lua_gettop (ls);
	log_entry ("%s():\n", __PRETTY_FUNCTION__);
	if (!lua_isstring (ls, 1)) {
		log_entry ("%s: Argument is not a string.\n",
			__PRETTY_FUNCTION__);
		return 0;
	}
	str = lua_tostring (ls, 1);

	lua_getglobal (ls, "alert");
	lua_pushstring (ls, "id");
	lua_gettable (ls, -2);
	n = lua_tonumber (ls, -1);
	a_id = (long long)n;

	result = dbi_conn_query (Manager.db_conn, "start transaction");
	if (result == NULL) {
		log_entry ("%s: Failed to execute start tx SQL.\n",
			__PRETTY_FUNCTION__);
		return 0;
	}
	dbi_result_free (result);

	result = dbi_conn_queryf (Manager.db_conn, 
				"insert into alert_data (ad_a_id,ad_timestamp,ad_type,ad_name,ad_data) values (%llu,NOW(),2,'SYSTEM','%s')",
				a_id,
				str);
	if (result == NULL) {
		log_entry ("%s: Failed to execute data SQL (rollback).\n",
			__PRETTY_FUNCTION__);
		result = dbi_conn_query (Manager.db_conn, "rollback");
		dbi_result_free (result);
		return 0;
	}
	dbi_result_free (result);

	result = dbi_conn_query (Manager.db_conn, "commit");
	if (result == NULL) {
		log_entry ("%s: Failed to execute commit tx SQL.\n",
			__PRETTY_FUNCTION__);
		return 0;
	}
	dbi_result_free (result);

	return 0;
}


int filter_list_item_compare (FilterListItem *a, FilterListItem *b)
{
	return strcmp (a->oid, b->oid);
}


int filter_match_test (FilterMatch *fm, Trap *trap)
{
	int		rc;
	int64_t		i64;
	float		f;
	int		ovector[PCRE_OVEC_ITEMS];
	TrapOid		stoid;
	TrapOid		*toid = NULL;
	SList		*litem = NULL;

	/*
	 * First, find the matching OID, by; PCRE, strstr() or name
	*/
	if (fm->flags & MATCH_FLAG_PCREOID) {
		rc = 0;
		litem = trap->list;
		while (litem != NULL) {
			toid = slist_get_data (litem);
			rc = pcre_exec (fm->pcreoid,
					NULL,
					toid->name,
					strlen (toid->name),
					0,
					0,
					ovector,
					PCRE_OVEC_ITEMS);
			if (rc > 0)
				break;
			litem = slist_get_next (litem);
		}
		if (litem == NULL || rc <= 0)
			return 0;

	} else if (fm->flags & MATCH_FLAG_STRSTR) {
		litem = trap->list;
		while (litem != NULL) {
			toid = slist_get_data (litem);
			if (strstr(toid->name, fm->oid))
				break;
			litem = slist_get_next (litem);
		}
		if (litem == NULL)
			return 0;

	} else {
		stoid.name = fm->oid;
		stoid.value = NULL;
		toid = avltree_find (trap->tree, &stoid);
	}
	if (toid == NULL)
		return 0;

	/*
	 * Second, test the OID's value against the filtermatch
	*/
	if (fm->valuetype == VALTYPE_INTEGER) {
		i64 = (int64_t)atoi (toid->value);
		switch (fm->comparator) {
			case COMP_EQUAL:
				if (i64 == fm->value.i64)
					return 1;
				break;
			case COMP_NOT_EQUAL:
				if (i64 != fm->value.i64)
					return 1;
				break;
			case COMP_GREATERTHAN:
				if (i64 > fm->value.i64)
					return 1;
				break;
			case COMP_LESSTHAN:
				if (i64 < fm->value.i64)
					return 1;
				break;
			case COMP_GREATERTHAN_EQUAL:
				if (i64 >= fm->value.i64)
					return 1;
				break;
			case COMP_LESSTHAN_EQUAL:
				if (i64 <= fm->value.i64)
					return 1;
				break;
			case COMP_REGEX:
				log_entry ("Can't regex an integer in filter match %llu.\n",
					fm->id);
				break;
			case COMP_NOT_REGEX:
				log_entry ("Can't regex an integer in filter match %llu.\n",
					fm->id);
				break;
			default:
				log_entry ("Invalid comparator in filter match %llu.\n",
					fm->id);
		}
	} else if (fm->valuetype == VALTYPE_FLOAT) {
		f = atof (toid->value);
		switch (fm->comparator) {
			case COMP_EQUAL:
				if (f == fm->value.f)
					return 1;
				break;
			case COMP_NOT_EQUAL:
				if (f != fm->value.f)
					return 1;
				break;
			case COMP_GREATERTHAN:
				if (f > fm->value.f)
					return 1;
				break;
			case COMP_LESSTHAN:
				if (f < fm->value.f)
					return 1;
				break;
			case COMP_GREATERTHAN_EQUAL:
				if (f >= fm->value.f)
					return 1;
				break;
			case COMP_LESSTHAN_EQUAL:
				if (f <= fm->value.f)
					return 1;
				break;
			case COMP_REGEX:
				log_entry ("Can't regex a float in filter match %llu.\n",
					fm->id);
				break;
			case COMP_NOT_REGEX:
				log_entry ("Can't regex a float in filter match %llu.\n",
					fm->id);
				break;
			default:
				log_entry ("Invalid comparator in filter match %llu.\n",
					fm->id);
		}
	} else if (fm->valuetype == VALTYPE_STRING) {
		switch (fm->comparator) {
			case COMP_EQUAL:
				rc = strcmp (toid->value, fm->value.s);
				if (rc == 0)
					return 1;
				break;
			case COMP_NOT_EQUAL:
				rc = strcmp (toid->value, fm->value.s);
				if (rc != 0)
					return 1;
				break;
			case COMP_GREATERTHAN:
				rc = strcmp (toid->value, fm->value.s);
				if (rc > 0)
					return 1;
				break;
			case COMP_LESSTHAN:
				rc = strcmp (toid->value, fm->value.s);
				if (rc < 0)
					return 1;
				break;
			case COMP_GREATERTHAN_EQUAL:
				rc = strcmp (toid->value, fm->value.s);
				if (rc >= 0)
					return 1;
				break;
			case COMP_LESSTHAN_EQUAL:
				rc = strcmp (toid->value, fm->value.s);
				if (rc <= 0)
					return 1;
				break;
			case COMP_REGEX:
				rc = pcre_exec (fm->pcrecode,
						NULL,
						toid->value,
						strlen (toid->value),
						0,
						0,
						ovector,
						PCRE_OVEC_ITEMS);
				if (rc > 0)
					return 1;
				break;
			case COMP_NOT_REGEX:
				rc = pcre_exec (fm->pcrecode,
						NULL,
						toid->value,
						strlen (toid->value),
						0,
						0,
						ovector,
						PCRE_OVEC_ITEMS);
				if (rc == -1)
					return 1;
				break;
			default:
				log_entry ("Invalid comparator in filter match %llu.\n",
					fm->id);
		}
	} else {
		log_entry ("Invalid valuetype in filter match %llu.\n",
			fm->id);
	}
	return 0;
}


int trapoid_compare (TrapOid *a, TrapOid *b)
{
	return strcmp (a->name, b->name);
}


int trapoid_destroy (TrapOid *a)
{
	if (a->name != NULL)
		free (a->name);
	if (a->value != NULL)
		free (a->value);
	free (a);
	return 0;
}


TrapOid *trapoid_tree_find (AVLTree *tree, char *name)
{
	TrapOid		sto;
	sto.name = name;
	sto.value = NULL;
	return avltree_find (tree, &sto);
}


int string_to_nvdata (NVData **nvdp,
			char *string,
			char *pairdelim,
			char nvdelim)
{
	int		rc;
	int		items;
	char		*p = NULL;
	char		*s = NULL;
	char		*nstr = NULL;
	char		*token = NULL;
	char		**nvpair = NULL;
	NVData		*nvd = NULL;
	nvd = calloc (1, sizeof (NVData));
	if (nvd == NULL) {
		return errno;
	}
	nstr = strdup (string);
	if (nstr == NULL) {
		free (nvd);
		return errno;
	}
	rc = avltree_new (&nvd->tree,
			(AVLNodeCompare)trapoid_compare,
			(AVLNodeDestroy)trapoid_destroy);
	if (rc != 0) {
		free (nstr);
		free (nvd);
		return rc;
	}
	for (s = nstr ; ; s = NULL) {
		token = strtok_r (s, pairdelim, &p);
		if (token == NULL)
			break;
		items = 2;
		rc = strnsplit (token,
				strlen (token),
				nvdelim,
				&nvpair,
				&items);
		if (rc != 0) {
			avltree_free (&nvd->tree);
			free (nstr);
			free (nvd);
			return rc;
		}
		avltree_insert (nvd->tree, nvpair);
	}
	free (nstr);
	*nvdp = nvd;
	return 0;
}


int handle_matching_trap (FilterRule *fr, Trap *trap)
{
	int		rc;
	int		flags;
	int		i;
	int		count;
	uint64_t	a_id;
	uint32_t	a_state;
	uint32_t	a_severity;
	uint32_t	a_hitcount;
	int		ovector[PCRE_OVEC_ITEMS];
	char		message[STRING_BUFFER_SIZE];;
	char		addlmsg[STRING_BUFFER_SIZE];;
	char		sql[SQL_BUFFER_SIZE];
	char		sqladd1[SQL_BUFFER_SIZE];
	char		sqladd2[SQL_BUFFER_SIZE];
	char		text[TRAP_BUFFER_SIZE];
	dbi_result	result;
	char		*object = NULL;
	char		*uniquestr = NULL;
	const char	*errmsg = NULL;
	TrapOid		*to = NULL;
	TrapOid		*ton = NULL;

	log_entry ("DEBUG: handle_matching_trap: tf_id=%llu\n", fr->id);
	log_entry ("DEBUG:  trap->trapoid='%s'\n", trap->trapoid);
	
	flags = 0;
	message[0] = 0;

	/*
	 * Handle pre-notification, if requested
	*/
	to = trapoid_tree_find (fr->actions->tree, "notify");
	if (to != NULL && to->value[0] != 0) {
		rc = emit_notification (fr, trap, to->value);
	}

	/*
	 * Drop this trap, if requested
	*/
	to = trapoid_tree_find (fr->actions->tree, "action");
	if (to != NULL && to->value[0] != 0) {
		if (!strcmp (to->value, "drop")) {
			log_entry ("DEBUG: Dropping trap.\n");
			return 0;
		} else if (!strcmp (to->value, "continue")) {
			flags |= TF_FLAG_CONTINUE;
		}
	}

	/*
	 * Get the objectname from the trap, or from the filter rule
	*/
	to = trapoid_tree_find (fr->actions->tree, "objectoid");
	if (to != NULL) {
		to = trapoid_tree_find (trap->tree, to->value);
		if (to != NULL) {
			object = to->value;
		}
	} else {
		object = fr->object;
	}
	log_entry ("DEBUG:  object = '%s'.\n", object);

	/*
	 * uniquestr_oid/uniquestr_regex  handling, if set in the filter
	*/
	if (fr->uniquestr_oid != NULL && fr->uniquestr_regex != NULL) {
		/* FIXME: need to be able to find the OID via regex */
		to = trapoid_tree_find (trap->tree, fr->uniquestr_oid);
		if (to != NULL) {
			/* Execute the PCRE */
			count = pcre_exec (fr->uniquestr_regex,
					NULL,
					to->value,
					strlen (to->value),
					0,
					0,
					ovector,
					PCRE_OVEC_ITEMS);
			/* Add the backrefs to the trap tree */
			i = 0;
			while (i < count) {
				/* FIXME: Check allocations */
				ton = calloc (1, sizeof (TrapOid));
				ton->name = calloc (1, UNIQUESTR_OID_SIZE);
				rc = snprintf (ton->name,
						UNIQUESTR_OID_SIZE,
						"unique.match.%d",
						i);
				rc = pcre_get_substring (to->value,
						ovector,
						count,
						i,
						(const char **)&ton->value);
				rc = avltree_insert (trap->tree, ton);
				rc = slist_item_add (&trap->list, ton);
				i++;
			}
		}
	}
	
	/*
	 * uniquestr handling, if set in the filter
	*/
	if (fr->uniquestr_parts != NULL) {
		/* build the uniquestr */
		rc = filter_rule_build_uniquestr (fr,
						trap,
						&uniquestr);
		/* TODO: extend DB field length and remove this */
		if (uniquestr != NULL && 
		    strlen (uniquestr) >= UNIQUESTR_FIELD_SIZE)
			uniquestr[UNIQUESTR_FIELD_SIZE-1] = 0;
	}
	// Use 'hostname-objectname' for the uniquestr value, if not already defined
	if (uniquestr == NULL) {
		uniquestr = calloc (1, UNIQUESTR_FIELD_SIZE);
		snprintf (uniquestr, UNIQUESTR_FIELD_SIZE, "%s-%s", trap->hostname, object);
	}
	log_entry ("DEBUG:  uniquestr = '%s'.\n", uniquestr);

	/*
	 * Build the message from any configued 'msgoid' or 'message' settings
	*/
	to = trapoid_tree_find (fr->actions->tree, "message");
	if (to != NULL && to->value != NULL) {
		strcat (message, to->value);
		strcat (message, "\n");
		
	}
	to = trapoid_tree_find (fr->actions->tree, "msgoid");
	if (to != NULL && to->value != NULL) {
		to = trapoid_tree_find (trap->tree, to->value);
		if (to != NULL && to->value != NULL) {
			strcat (message, to->value);
			strcat (message, "\n");
		}
	}
	log_entry ("DEBUG:  message = '%s'.\n", message);

	/*
	 * If this filter is not flagged as always unique,
	 * search for an existing alert record
	*/
	a_id = 0;
	sqladd1[0] = 0;
	if (uniquestr != NULL) {
		rc = snprintf (sqladd1,
				SQL_BUFFER_SIZE,
				"and a_uniquestr = '%s'",
				uniquestr);
	}
	sqladd2[0] = 0;
	if (fr->uniqueid > 0) {
		rc = snprintf (sqladd2,
				SQL_BUFFER_SIZE,
				"and a_uniqueid = %lu",
				(long unsigned int)fr->uniqueid);
	}
	rc = snprintf (sql,
			SQL_BUFFER_SIZE,
			"select a_id,a_state,a_hitcount,a_severity from alert where a_hostname = '%s' and a_object = '%s' and a_class = %d and a_state not in (4,5) %s %s",
				trap->hostname,
				object,
				fr->class,
				sqladd1,
				sqladd2);
	log_entry ("DEBUG:  alert search = '%s'.\n", sql);
	result = dbi_conn_query (Manager.db_conn, sql);
	if (result == NULL) {
		rc = dbi_conn_error (Manager.db_conn, &errmsg);
		log_entry ("Failed to execute alert search SQL.\n");
		log_entry ("dbierr: %s\n", errmsg);
		if (uniquestr != NULL)
			free (uniquestr);
		return RC_DBERROR;
	}
	if (dbi_result_next_row (result)) {
		a_id = dbi_result_get_ulonglong (result, "a_id");
		a_state = dbi_result_get_uint (result, "a_state");
		a_hitcount = dbi_result_get_uint (result, "a_hitcount");
		a_severity = dbi_result_get_uint (result, "a_severity");
	}
	dbi_result_free (result);

	if (a_id > 0) {
		/*
		 * A matching alert record was found
		*/
		sqladd1[0] = 0;
		sqladd2[0] = 0;
		log_entry ("DEBUG:  search found existing(matching) alert a_id=%d.\n", a_id);
		/*
		log_entry ("DEBUG: fr->id  = '%d'.\n", fr->id);
		log_entry ("DEBUG: fr->class  = '%d'.\n", fr->class);
		log_entry ("DEBUG: fr->severity  = '%d'.\n", fr->severity);
		log_entry ("DEBUG: fr->markrecover  = '%d'.\n", fr->markrecover);
		log_entry ("DEBUG: a_id  = '%d'.\n", a_id);
		log_entry ("DEBUG: a_state  = '%d'.\n", a_state);
		log_entry ("DEBUG: a_hitcount  = '%d'.\n", a_hitcount);
		log_entry ("DEBUG: a_severity  = '%d'.\n", a_severity);
		*/
		to = trapoid_tree_find (fr->actions->tree, "appendevent");
		if (to != NULL) {
			i = atoi (to->value);
			if (i >= 1) {
				/* Only append if not a 'clearing' event */
				if (a_state >= 1 &&
				    a_state <= 3 &&
				    fr->markrecover != 1) {
					rc = alert_add_data (a_id, i, message);
				}
			}
		}
		if (a_state == 3) {
			/* If Ack'd, just increment hitcount */
			a_hitcount++;
		} else {
			/* If not Ack'd, and it's a recover filter */
			if (fr->markrecover == 1) {
				log_entry ("DEBUG: Changing state for recovery filter match. alert a_id=%d.\n", a_id);
				a_state = 4;
				rc = alert_add_data (a_id, a_state, message);
			} else {
				/* Mark as a Dup(icate) */
				a_state = 2;
				a_hitcount++;
			}
		}

		if (fr->markrecover != 1 && fr->severity != a_severity) {
			/* Non-recovery, severity changed */
			rc = snprintf (sqladd1,
					SQL_BUFFER_SIZE,
					", a_severity = %lu",
					(long unsigned int)fr->severity);
			rc = snprintf (addlmsg,
					STRING_BUFFER_SIZE,
					"The severity changed from %u to %u",
					a_severity,
					fr->severity);
			rc = alert_add_data (a_id, a_state, addlmsg);
			a_severity = fr->severity;
		}

		rc = snprintf (sql,
				SQL_BUFFER_SIZE,
				"update alert set a_lasttime = from_unixtime(%lu), a_state = %lu, a_hitcount = %lu, a_message = '%s' %s %s where a_id = %llu",
				(long unsigned int)trap->timestamp,
				(long unsigned int)a_state,
				(long unsigned int)a_hitcount,
				message,
				sqladd1,
				sqladd2,
				(long long unsigned int)a_id);
		log_entry ("DEBUG:  alert update = '%s'.\n", sql);
		result = dbi_conn_query (Manager.db_conn, sql);
		if (result == NULL) {
			rc = dbi_conn_error (Manager.db_conn, &errmsg);
			log_entry ("Failed to execute alert update SQL.\n");
			log_entry ("dbierr: %s\n", errmsg);
			if (uniquestr != NULL)
				free (uniquestr);
			return RC_DBERROR;
		}
		dbi_result_free (result);

	} else {
		/*
		 * No matching, existing alert record was found
		*/

		log_entry ("DEBUG:  No existing (matching) alert found.\n");
		if (fr->markrecover == 1) {
			/* Filter is markrecover, but no matching alert
			   record was found. */
			log_entry ("Warning: The following 'recover' event was processed, but no matching alert record was found:\n");
			log_entry ("trap_filter: id=%llu name='%s'\n",
				fr->id,
				fr->name);
			log_entry ("Event: hostname='%s' message='%s'\n",
				trap->hostname,
				message);
			if (uniquestr != NULL)
				free (uniquestr);
			if (flags & TF_FLAG_CONTINUE)
				return RC_CONTINUE;
			return 0;
		}

		/*
		 * insert the new alert record
		*/
		rc = snprintf (sql,
				SQL_BUFFER_SIZE,
				"insert into alert (%s) values (from_unixtime(%lu),from_unixtime(%lu),%lu,%lu,%lu,%lu,'%s','%s','%s','%s',%lu)",
				"a_alerttime,a_lasttime,a_class,a_severity,a_state,a_uniqueid,a_hostname,a_uniquestr,a_object,a_message,a_hitcount",
				(long unsigned int)trap->timestamp,
				(long unsigned int)trap->timestamp,
				(long unsigned int)fr->class,
				(long unsigned int)fr->severity,
				(long unsigned int)1,
				(long unsigned int)fr->uniqueid,
				trap->hostname,
				uniquestr,
				object,
				message,
				(long unsigned int)1);
		log_entry ("DEBUG:  alert insert = '%s'.\n", sql);
		result = dbi_conn_query (Manager.db_conn, sql);
		if (result != NULL) {
			a_id = dbi_conn_sequence_last (Manager.db_conn, NULL);
		} else {
			rc = dbi_conn_error (Manager.db_conn, &errmsg);
			log_entry ("Failed to execute alert insert SQL.\n");
			log_entry ("dbierr: %s\n", errmsg);
			if (uniquestr != NULL)
				free (uniquestr);
			return RC_DBERROR;
		}
		dbi_result_free (result);

		/*
		 * If the alert was successfully added, insert the
		 * alert_data records
		*/
		if (a_id > 0) {
			//strcpy (text, "TODO: Add the trap content here.");
			rc = trap_dump_to_string (trap, text, TRAP_BUFFER_SIZE);
			/* Insert the trap content record */
			rc = alert_add_data (a_id, 6, text);
			/* Inert the 'event' record */
			rc = alert_add_data (a_id, 1, message);
		}
		/* commit() */

		/*
		 * if requested, send a notification for the new alert
		*/
		to = trapoid_tree_find (fr->actions->tree, "notifynew");
		if (to != NULL && to->value[0] != 0) {
			rc = emit_notification (fr, trap, to->value);
		}

	}
	if (uniquestr != NULL)
		free (uniquestr);
	if (flags & TF_FLAG_CONTINUE)
		return RC_CONTINUE;
	return 0;
}


int emit_notification (FilterRule *fr, Trap *trap, char *notifylist)
{
	log_entry ("DEBUG: Notifying '%s'\n", notifylist);
	return 0;
}


int parse_string_to_template (SList **slp, char *string)
{
	int		rc;
	char		*str = NULL;
	char		*p = NULL;
	char		*tokp = NULL;
	char		*token = NULL;
	SList		*list = NULL;
	StrPart		*sp = NULL;

	if (slp == NULL || string == NULL)
		return EINVAL;
	*slp = NULL;
	str = strdup (string);
	if (str == NULL)
		return errno;
	p = str;
	tokp = NULL;
	while (1) {
		token = strtok_r (p, "{", &tokp);
		if (token == NULL)
			break;
		//log_entry ("DEBUG: token='%s'\n", token);
		sp = calloc (1, sizeof (StrPart));
		if (sp == NULL) {
			rc = errno;
			slist_free (&list);
			free (str);
			return rc;
		}
		sp->type = STRPART_STRING;
		sp->string = strdup (token);
		rc = slist_item_add (&list, sp);
		p = NULL;
		token = strtok_r (p, "}", &tokp); 
		if (token == NULL)
			break;
		//log_entry ("DEBUG: token='%s'\n", token);
		sp = calloc (1, sizeof (StrPart));
		if (sp == NULL) {
			rc = errno;
			slist_free (&list);
			free (str);
			return rc;
		}
		sp->type = STRPART_BACKREF;
		sp->string = strdup (token);
		rc = slist_item_add (&list, sp);
	}
	sp = calloc (1, sizeof (StrPart));
	if (sp == NULL) {
		rc = errno;
		slist_free (&list);
		free (str);
		return rc;
	}
	sp->type = STRPART_STRING;
	sp->string = strdup (tokp);
	rc = slist_item_add (&list, sp);
	free (str);
	*slp = list;
	return 0;
}


int filter_rule_build_uniquestr (FilterRule *fr, Trap *trap, char **stringp)
{
	int		len;
	int		bytes;
	char		*p = NULL;
	char		*str = NULL;
	char		*value = NULL;
	SList		*litem = NULL;
	StrPart		*sp = NULL;
	TrapOid		*to = NULL;

	str = malloc (STRING_BUFFER_SIZE);
	if (str == NULL)
		return errno;
	bytes = STRING_BUFFER_SIZE;

	p = str;
	litem = fr->uniquestr_parts;
	while (litem != NULL) {
		sp = (StrPart *)slist_get_data (litem);
		if (sp->type == STRPART_BACKREF) {
			to = trapoid_tree_find (trap->tree, sp->string);
			if (to != NULL && to->value != NULL)
				value = to->value;
			else
				log_entry ("DEBUG:  Unset OID '%s' while building uniquestr\n",
					fr->id,
					sp->string);
		} else {
			value = sp->string;
		}
		len = strlen (value);
		if (len >= bytes) {
			free (str);
			return ENOMEM;
		}
		memcpy (p, value, len);
		p += len;
		bytes -= len;
		litem = slist_get_next (litem);
	}
	*p = 0;
	*stringp = str;
	return 0;
}


int alert_add_data (uint64_t a_id, uint32_t type, char *message)
{
	int		rc;
	dbi_result	result;
	char		sql[SQL_BUFFER_SIZE];
	const char	*errmsg = NULL;
	rc = snprintf (sql,
			SQL_BUFFER_SIZE,
			"insert into alert_data (%s) values (%llu,from_unixtime(%lu),%lu,'%s','%s')",
			"ad_a_id,ad_timestamp,ad_type,ad_name,ad_data",
			(long long unsigned int)a_id,
			time (NULL),
			(long unsigned int)type,
			"SYSTEM",
			message);
	result = dbi_conn_query (Manager.db_conn, sql);
	if (result == NULL) {
		rc = dbi_conn_error (Manager.db_conn, &errmsg);
		log_entry ("Failed to execute alert_data insert SQL.\n");
		log_entry ("dbierr: %s\n", errmsg);
		return RC_DBERROR;
	}
	dbi_result_free (result);
	return 0;
}



int load_configuration_file (char *file)
{
	int		rc;
	char		*str;
	Configuration	*c;
	ConfigSearch	*cs;
	ConfigItem	*ci;

	configuration_init(NULL);
	rc = configuration_new (&c);
	if (rc != ESUCCESS)
		return rc;
	
	rc = configuration_load_file (c, file);
	if (rc != ESUCCESS) {
		configuration_free (&c);
		return rc;
	}

	rc = configuration_search_new (c, &cs, NULL);

	// Logging properties
	rc = configuration_search_begin (cs,
				"/netinfo/netinfofilterd/configuration/logging");
	rc = configuration_search_next_item (cs, &ci);
	if (rc != ESUCCESS) {
		configuration_search_end (cs);
		configuration_search_free (&cs);
		return rc;
	}
	str = configuration_item_get_attribute (ci, "file");
	strncpy(Manager.log_file, str, PATH_SIZE);
	configuration_search_end (cs);

	rc = configuration_search_begin (cs,
				"/netinfo/netinfofilterd/configuration/logging/options");
	rc = configuration_search_next_item (cs, &ci);
	if (rc != ESUCCESS) {
		configuration_search_end (cs);
		configuration_search_free (&cs);
		return rc;
	}
	str = configuration_item_get_attribute (ci, "flush_log_entries");
	if (!strcmp(str, "1"))
		Manager.flags |= FLAG_FLUSH_LOG_ENTRIES;
	str = configuration_item_get_attribute (ci, "log_unhandled_traps");
	if (!strcmp(str, "1"))
		Manager.flags |= FLAG_LOG_UNHANDLED_TRAPS;
	configuration_search_end (cs);

	// Queue location properties
	rc = configuration_search_begin (cs,
				"/netinfo/netinfofilterd/configuration/queue");
	rc = configuration_search_next_item (cs, &ci);
	if (rc != ESUCCESS) {
		configuration_search_end (cs);
		configuration_search_free (&cs);
		return rc;
	}
	str = configuration_item_get_attribute (ci, "path");
	Manager.pickupdir_path = str;
	configuration_search_end (cs);

	// DB properties
	rc = configuration_search_begin (cs,
				"/netinfo/netinfofilterd/configuration/database");
	rc = configuration_search_next_item (cs, &ci);
	if (rc != ESUCCESS) {
		configuration_search_end (cs);
		configuration_search_free (&cs);
		return rc;
	}
	str = configuration_item_get_attribute (ci, "driver");
	Manager.db_driver = str;
	str = configuration_item_get_attribute (ci, "hostname");
	Manager.db_hostname = str;
	str = configuration_item_get_attribute (ci, "port");
	Manager.db_port = atoi(str);
	free (str);
	str = configuration_item_get_attribute (ci, "dbname");
	Manager.db_dbname = str;
	str = configuration_item_get_attribute (ci, "username");
	Manager.db_username = str;
	str = configuration_item_get_attribute (ci, "password");
	Manager.db_password = str;
	configuration_search_end (cs);

	configuration_search_free (&cs);

	return 0;
}


int trap_dump_to_string (Trap *trap, char *str, uint32_t strlen)
{
	int		rc;
	int		rem;
	char		*p = str;
	SList		*litem = NULL;
	TrapOid		*toid = NULL;

	if (trap == NULL || str == NULL)
		return EINVAL;
	if (strlen < 128)
		return ENOSPC;

	rem = strlen-1;
	for (litem=trap->list ; litem != NULL ; litem=slist_get_next(litem)) {
		toid = slist_get_data(litem);
		rc = snprintf (p, rem, "%s: %s\n", toid->name, toid->value);
		if (rc < 0)
			return 1;
		if (rc > rem)
			break;
		rem -= rc;
		p += rc;
	}
	str[strlen-1] = '\0';
	return 0;
}


void usage(void)
{
	printf("\nnetinfofilterd - Netinfo Trap Filtering Server\n");
	printf(" Options:\n");
	printf("   -D              Netinfofilterd will not daemonize itself.\n");
	printf("   -F              Log entries will be flushed to disk immediately.\n");
	printf("   -f configfile   Specifies the path to the configuration file to use.\n");
	printf("   -h              This help screen.\n");
	printf("   -l logfile      Specifies the path to the log file to use.\n");
	printf("   -V              Display version information and exit.\n\n");
	return;
}

